/*
 * This file is part of a Lucent implementation of kermit
 * 
 * This software is provided in the hope that it will be 
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * It was developed by Lucent technologies and is being made available to
 * Columbia University for their public or proprietary use.
 *
 * Author: Andrew Ellis
 * Copyright (C) 2003 Lucent Technoligies, Inc.
 *
 * The Kermit protocol is copyrighted by Columbia University
 *
 * Lucent retains the right to use, modify and further develop this software
 * and any future versions of this software.
 */

package com.lucent.kermit;

import java.awt.Window;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;

import java.text.SimpleDateFormat;

import java.util.Date;
import java.util.Stack;
import java.util.EventObject;

/**
 * This file implements a filter for kermit
 */

public class Kermit {

  static public final int OFF = 0;		// Currently OFF
  static public final int RECV_PENDING = 1;	// Currently Idle - waiting
  						//		on sendInit
  static public final int RECV_FH_PENDING = 2;	// Expecting file header
  static public final int RECV_DATA_PENDING = 3;// Expecting file header
  static public final int RECEIVING_DATA = 4;	// Currently receiving data
  static public final int SEND_PENDING = 10;	// Currently sending
  static public final int FH_PENDING = 11;	// Waiting for ack frm sendInit
  static public final int ATTR_PENDING = 12;	// Waiting for ack frm file hdr
  static public final int DATA_PENDING = 13;	// Waiting for ack to send
  						//                data pkt
  static public final int EOF_PENDING = 14;	// Waiting for ack to snd
						//                last data pkt
  static public final int EOT_PENDING = 15;	// Waiting for ack from eof
  static public final int OFF_PENDING = 20;	// Waiting for ack to send data

  static final int CAP_LP = 2;		// Long packet capability
  static final int CAP_SW = 4;		// Sliding windows capability
  static final int CAP_AT = 8;		// Attribute packet capability
  static final int CAP_RS = 16;		// Resend capability
  static final int CAP_LS = 32;		// Locking shift capability
  static final int MAX_LONG_PACKET_LENGTH = 1024;
  					// Maximum packet length with long pkts
  static final int MAX_PACKET_LENGTH = 94;
					// Maximum packet length

  final static SimpleDateFormat		// Used for attributes
    dateFormatter = new SimpleDateFormat( "yyyyMMdd kk:mm:ss" );

  static int debug = 0;

  // CRC tables
  // CRC generation table A
  protected final long[] crcTableA = {
    0,		010201,		020402,		030603,
    041004,	051205,		061406,		071607,
    0102010,	0112211,	0122412,	0132613,
    0143014,	0153215,	0163416,	0173617
  };

  // CRC generation table B
  protected final long[] crcTableB = {
    0,		010611,		021422,		031233,
    043044,	053655,		062466,		072277,
    0106110,	0116701,	0127532,	0137323,
    0145154,	0155745,	0164576,	0174367
  };

  protected KermitController	kController = null;	// caller

  protected int		kState = OFF;	// state of kermit
  protected Stack	kFileStack;	// stack of files to send
  protected KermitPkt	kPkt = null;	// packet we are now receiving
  protected byte[]	kLastPkt = null;// last packet we sent
  protected File	kFile = null;	// local file open for reading/writing
  protected int		kSeq;		/* sending: the next number we
					 * 	    are going to send.
					 * recving: the seqNo that came in the
					 *          last good packet
					 */
  protected int		lastAckedSeq;	// the seq num of the last ack we got
  protected String	kRemoteName = null;	// name of remote file

  protected BufferedOutputStream	kFileOutputStream;
  protected BufferedInputStream		kFileInputStream;
  protected int				kByteProgress;
  protected int				kFileSize;
  protected String			kStatus;

  protected byte[] headerBytes = null;	// collection of header bytes
  protected int headerByteIndex = 0;
  protected int lastHeaderByteIndex = 0;

  protected int maxBadPacketCount;	// Max # of timeouts allowed
  protected int curBadPacketCount;	// Current # of consecutive timeouts

  // If the parity state is non-zero, we need to use binary prefixes
  protected int parity = 0;		// Parity state.

  // These flags represent the states as agreed to by both sides of the pipe
  protected byte	sendSOH;	// Start of header char
  protected int		sendMaxl;	// maximum length of kermit packets
  protected int		sendTime;	// Time to wait before timeout/NACK
  protected int		sendNpad;	// Do we use padding - not supported
  protected char	sendPadc;	// padding character - not supported
  protected char	sendEol;	// Sent after checksum
  protected char	sendQctl;	// Control-shift prefix
  protected char	sendQbinChar;	// Binary-shift prefix
  protected boolean	sendQbinFlag;	// Binary-shift state
  protected int		sendChkt;	// Checksum type (1, 2, or 3 )
  protected char	sendRptChar;	// Repeat prefix
  protected boolean	sendRptFlag;	// false;
  protected byte	sendCapas;	// Capabilities mask
  protected int		sendLongMaxl;	// max length of long kermit packets

  // default values for the above
  static protected byte		defSendSOH = 1;
  static protected int		defSendMaxl = MAX_PACKET_LENGTH;
  static protected int		defSendTime = 5;
  static protected int		defSendNpad = 0;
  static protected char		defSendPadc = '@';
  static protected char		defSendEol = (char)toByte( '\r' );
  static protected char		defSendQctl = '#';
  static protected char		defSendQbinChar = '&';
  static protected boolean	defSendQbinFlag = false;
  static protected int		defSendChkt = 1;
  static protected char		defSendRptChar = '~';
  static protected boolean	defSendRptFlag = false;
  static protected byte		defSendCapas = 0|CAP_AT|CAP_LP;
  static protected int		defSendLongMaxl = MAX_LONG_PACKET_LENGTH;

  /**
   * @param KermitController controller
   *
   */
  public Kermit( KermitController controller ) {
    // just initialize values and wait.  Either the controller will tell us
    // what to do next or we will receive a send_init from the other side.
    kController = controller;
    setInitValues();
  }

  /** Conversion routines for header information. */
  static protected int unByte( byte b ) { return b - ' '; }
  static protected byte toByte( int i ) { return (byte)((i + ' ') & 0177); }

  /** Set the default character for start of packet information
   * @param ascii The ascii character used to start kermit packets
   */
  static public void setDefaultSOH( int ascii ) {
    if( ascii < 0 || ascii > 127 ) return;
    defSendSOH = (byte)ascii;
  }

  /** Sets the default value for the maximum packet length that this
   * application can send or receive.
   * @param length The length that can be sent or received
   */
  static public void setDefaultMaxPacketLength( int length ) {
    if( length < 60 ) length = 60;
    if( length > MAX_PACKET_LENGTH ) {
      defSendMaxl = MAX_PACKET_LENGTH;
      defSendLongMaxl = length;
    } else {
      defSendMaxl = defSendLongMaxl = length;
    }
  }

  /** Sets the default character that will be appended to each kermit packet.
   * @param ascii The character to be appended to each kermit packet
   */
  static public void setDefaultEndOfLine( int ascii ) {
    if( ascii < 0 || ascii > 95 ) return;
    defSendEol = (char)toByte( ascii );
  }

  /** Sets the default character that will be used as the control prefix
   * @param ascii the character to be used as the control prefix
   */
  static public void setDefaultControlPrefix( int ascii ) {
    if( ascii < 32 || ascii > 126 ) return;
    defSendQctl = (char) ascii;
  }

  /** Sets the default character that will be used as the binary prefix
   * @param ascii the character to be used as the binary prefix
   */
  static public void setDefaultBinaryPrefix( int ascii ) {
    if( ascii < 32 || ascii > 126 ) return;
    defSendQbinChar = (char) ascii;
  }

  /** Sets the default method for packet verification
   * @param type Should be 1, 2, or 3 for the number of characters used
   */
  static public void setDefaultCheckSum( int type ) {
    if( type > 0 && type < 4 ) defSendChkt = type;
  }

  /** Starts the process of sending this file via kermit.
   *  This should only be called when kermit is OFF
   * @param file The file to be sent out.
   */
  public void sendFile( File file ) {
    addFile( file );
    sendNextFile();
  }

  /** Starts the process of sending files to kermit.
   *  This should only be called when kermit is OFF
   * @param fileArray An array of file objects to be sent off.
   */
  public void sendFiles( File[] fileArray ) {
    for( int i = fileArray.length-1; i >= 0; i-- ) addFile( fileArray[i] );
    sendNextFile();
  }

  // pushes the file on the stack of files to be sent out
  protected void addFile( File file ) { kFileStack.push( file ); }

  // Get the first, or next, file going here
  protected void sendNextFile() {

    /* If we are RECV_PENDING then we have not started sending files yet.
     * We perform a send_init packet and return.  This method will be called
     * again as soon as we get the ACK for the send_init packet.
     */

    if( kState == RECV_PENDING ) { /* not yet initialized */
      sendInit();
      return;
    }

    /* The assumption is that we have just gotten the send_init ack
     * or the eof packet from the last file we sent.
     *
     * It is now time to send out the next file header packet or end
     * the transmission if there is no next file
     */
    
    if( kFileStack.size() <= 0 ) { /* no more files */
      sendEOT();
      return;
    }

    // Get the next file name 
    kFile = (File)kFileStack.pop();
    kRemoteName = kFile.getName();
    kFileSize = (int)kFile.length();
    kByteProgress = 0;

    // open the file for reading
    try {
      kFileInputStream = new BufferedInputStream(new FileInputStream( kFile ));
    } catch( Exception e ) {
      kStatus = "Could not open file "+e.getMessage();
      shutdown();
      return;
    }

    // create and send the header packet
    byte[] packet = kRemoteName.getBytes();

    // Set our state so we know what to do when they respond to the FH pkt
    kState = DATA_PENDING;
    if( (sendCapas & CAP_AT) > 0 ) kState = ATTR_PENDING;

    // send the file header packet
    kStatus = "Sending file header...";
    sendPacket( (byte)'F', packet );
  }

  // Kics off the the send process by sending the send_init packet
  protected void sendInit() {
    kState = SEND_PENDING; // we are about to send

    byte[] packet = createSendInitPacket();
    //    dumpSendInit( packet );

    int sendChktSave = sendChkt;
    sendChkt = 1; // always use this when sending an 'S' Packet
    sendPacket( (byte)'S', packet );
    sendChkt = sendChktSave;

    kState = FH_PENDING;

    kStatus = "Initialized.  Waiting for response...";
  }

  protected void setInitValues() {
    // initialize the fields of the send init packet.  These should work for
    // either direction
    sendSOH = defSendSOH;
    sendMaxl = defSendMaxl;
    sendTime = defSendTime;
    sendNpad = defSendNpad;
    sendPadc = defSendPadc;
    sendEol = defSendEol;
    sendQctl = defSendQctl;
    sendQbinChar = defSendQbinChar;
    sendChkt = defSendChkt;
    sendRptChar = defSendRptChar;	sendRptFlag = defSendRptFlag;
    sendCapas = defSendCapas;
    sendLongMaxl = defSendLongMaxl;

    // packet intiialization
    kPkt = null;		// no packet received yet.
    kSeq = 0;
    lastAckedSeq = -1;
    curBadPacketCount = 0;
    maxBadPacketCount = 5;	// currently, this is fixed

    kState = RECV_PENDING;	// we are now waiting for a send_init
    kFileStack = new Stack();
    kByteProgress = 0;
    kFileSize = 0;
    kStatus = "waiting...";
  }

  /** Set the parity of the current kermit.  The calling application should
   * leave this alone unless the parity of the connection is not 'NONE'.
   * 
   * Since this is called only for the purposes of allowing the 8th bit to
   * be used by the connection, the parityValue should be 0 for 'NONE' or
   * any other number to signify that parity is set to something other than
   * 'NONE'.
   */
  public void setParity( int parityValue ) {
    parity = parityValue;
  }

  boolean isBadPacketCountTooHigh( String problem ) {
    if( curBadPacketCount++ >= maxBadPacketCount ) {
      sendError( problem+": Too many retries" );
      return true;
    }
    kStatus = "Timeout: retrying("+curBadPacketCount+")";
    return false;
  }



  /** This should be called if the calling application has waited too long
   * for a kermit packet.  Kermit will take it from there.
   */
  public void timeout() {
    if( isBadPacketCountTooHigh( "Timeout" ) ) return;

    // The read timed out.  Do the right thing with NAKs
    if( kState == RECV_PENDING ) {
      // If we are receiving send_init, then nak the expected seq
      sendNak( toByte( 0 ) );
    }
    else if( kState < SEND_PENDING ) {
      // If we are receiving, then nak the expected seq
      sendNak( toByte( getNextSeq( kSeq ) ) );
    }
    else {
      // If we are sending, then resend the last packet
      writePacket( kLastPkt );
    }
  }

  /** This is called when the calling application receives a packet while
   * running kermit.  The packet may contain one whole kermit packet,
   * or a piece of a kermit packet.  Multiple packets won't work since
   * we did not implement sliding windows and can only send one packet at
   * a time anyway.
   * @param b The packet specified as a byte array.
   * @param n The number of bytes in the byte array that are part of the packet
   */
  public void kermitGotPacket( byte[] b, int n ) {
    for( int i = 0; i < n; ) {

      byte firstByte = b[i++];
      // The SOH should never show up unless we starting a new kermit packet
      if( firstByte == sendSOH ) { // SOH
	// Since this is the first packet, we expect it to be at the
	// beginning of a large packet.  The Java IO stream we are reading from
	// does not guarantee that we will receive our 'kermit' packet in one
	// read.  Nor does it guarantee that we will receive packets containing
	// only one kermit packet.

	// Therefore we have to leave things in a state as if the packet could
	// be split at any point.

	// We will collect header bytes until we have enough to make a header.

	if( kPkt != null ) {	// last packet was never completed
	  dbg( 0, "Last packet was not completed" );
	  // we need to nak the last packet - not this one
	  int lastSeq = (kPkt==null)?0:kPkt.getSeq();
	  sendNak( toByte( lastSeq ) );
	  kPkt = null;
	  headerBytes = null;
	  return; // I don't care what else my peer is sending
	}

	// start collecting header bytes
	headerBytes = new byte[10];
	headerByteIndex = 0;
	headerBytes[headerByteIndex++] = firstByte;
	continue;
      } // done with SOH

      if( headerBytes != null ) { // We are collecting header bytes
	if( headerByteIndex == 1 ) { // This should be the length
	  headerBytes[headerByteIndex++] = firstByte;
	  lastHeaderByteIndex = 3;
	  if( unByte( firstByte ) == 0 ) // long packet header
	    lastHeaderByteIndex = 6;
	} else {
	  // another header Byte - collect it
	  headerBytes[headerByteIndex++] = firstByte;

	  if( headerByteIndex > lastHeaderByteIndex ) {
	    // This was the last byte in the header.  Create the kermit pkt
	    // and start adding bytes to it.
	    try {
	      kPkt = createKermitPktFromHeader( headerBytes );
	    } catch( Exception e ) {
	      // since this packet is bad, we don't care about the rest
	      kStatus = "Bad header:" + e.getMessage();
	      sendNak( headerBytes[2] );
	      return;
	    }
	    headerBytes = null;
	  }
	}
	continue; // done with this header byte - get the next byte
      }

      if( kPkt == null ) {
	// there was an extra byte here - probably a trailing byte
	dbg( 1, "Extra byte:" + firstByte );
	continue;
      }

      kPkt.addByte( firstByte );

      if( kPkt.isComplete() )
	processKermitPacket();	// sets kermitPkt to null

      if( kState == OFF ) return; // we quit
    }
  }

  /** Creates a kermit packet structure that includes the header information */
  protected KermitPkt createKermitPktFromHeader( byte[] header )
    throws Exception {
    // 0 = SOH
    // 1 = LEN
    // 2 = SEQ
    // 3 = TYPE
    // and if length is 0
    // 4 = LENGTH - first byte
    // 5 = LENGTH - second byte
    // 6 = check sum for header

    int length = unByte( header[1] );
    if( length == 0 ) {
      // length = data packet length + chksumlen
      length = unByte( header[4] ) * 95;	// big part of length
      length += unByte( header[5] );		// little part of length
      byte chk = header[6];

      // check if the header's valid - always use type 1 checksum
      if( chk != toByte( checksum1( header, 1, 6 ) ) ) {
	String msg = "Bad checksum for header:" +
	  unByte( chk )+"/"+checksum1(header, 1, 6 );
	throw new Exception( msg );
      }
    }
    return new KermitPkt( unByte(header[2]), header[3], length, header );
  }

  /** Once a complete packet has been collected, process the packet. */
  void processKermitPacket() {
    byte type = kPkt.getType();
    dbg( 1, "Rcvd pkt:"
	 + "len real/decl:("+kPkt.getSize()+"/"+kPkt.getLength()+"),"
	 + "seq("+kPkt.getSeq()+"),"+"type("+(char)type+")" );

    if( !kPkt.isPacketOk() ) {
      dbg( 0, "Kermit Error: bad checksum" );
      sendNak( toByte( kPkt.getSeq() ) );
      kPkt = null;
      return;
    }

    // we will need to start responding to nacks
    if( type == 'Y' || type == 'N' ) {
    } else {
      kSeq = kPkt.getSeq();
    }

    if( type == 'Y' ) {
      handleAck( kPkt );
    }
    else if( type == 'D' ) {
      handleData( kPkt );
    }
    else if( type == 'F' ) {
      handleFileHeader( kPkt );
    }
    else if( type == 'S' ) {
      if( kState != RECV_PENDING ) {
	kStatus = "sendInit packet received when not in receive_init mode";
      } else {
	handleSendInit( kPkt );
	byte[] responsePkt = createSendInitPacket();
	//	dumpSendInit( responsePkt );

	int sendChktSave = sendChkt;
	sendChkt = 1; // always use this when sending an 'S' Packet
	sendPacket( (byte)'Y', responsePkt );
	sendChkt = sendChktSave;

	kStatus = "Waiting for file information...";
	kState = RECV_FH_PENDING;
      }
    }
    else if( type == 'N' ) {
      handleNak( kPkt );
    }
    else if( type == 'Z' ) {
      handleEOF( kPkt );
    }
    else if( type == 'B' ) {
      handleEndOfTransmission( kPkt );
    }
    else if( type == 'E' ) {
      handleError( kPkt );
    }
    else if( type == 'A' ) {
      handleAttributes( kPkt );
    }
    kPkt = null;
  }

  // 6 bit check sum
  protected int checksum1( byte[] packet, int start, int end ) {
    int checkSum = checksum2( packet, start, end );
    int checkByte = (byte)((checkSum + ((checkSum & 0300) >> 6)) & 077);
    return checkByte;
  }

  // 12 bit check sum
  protected int checksum2( byte[] packet, int start, int end ) {
    int checkSum = 0;
    // add up each byte
    for( int i = start; i < end; i++ ) checkSum += packet[i];
    return( checkSum & 07777 );
  }

  // 12 bit check sum
  protected int checksum3( byte[] packet, int start, int end ) {
    long c = 0;
    long crc = 0;
    // add up each byte
    for( int i = start; i < end; i++ ) {
      c = crc ^ (long)packet[i];
      crc =
	(crc >> 8)
	^ (crcTableA[ (int)(c & 0xf0) >> 4 ]
	   ^ crcTableB[(int)(c & 0x0f)]);
    }
    return( (int)(crc & 0xffff) );
  }

  // this is only used for debugging
  void dumpSendInit( byte[] packet ) {
    String s = "";
    // process the information in the packet and send a new packet back.
    int i = 0;
    if( packet[0] == (byte)1 ) i = 4; // we got the whole packet
    if( i < packet.length ) s += "maxl("+unByte( packet[i++] )+"),";
    if( i < packet.length ) s += "time("+unByte( packet[i++] )+"),";
    if( i < packet.length ) s += "npad("+unByte( packet[i++] )+"),";
    if( i < packet.length ) s += "padc("+(char)packet[i++]+"),";
    if( i < packet.length ) s += "eol("+(char)packet[i++]+"),";
    if( i < packet.length ) s += "qctl("+(char)packet[i++]+"),";
    if( i < packet.length ) s += "qbin("+(char)packet[i++]+"),";
    if( i < packet.length ) s += "chkt("+unByte( packet[i++] )+"),";
    if( i < packet.length ) s += "rpt("+(char)packet[i++]+"),";
    if( i < packet.length ) s += "capas(0"
		      +Integer.toOctalString((int)packet[i++]&0377)+")";
    dbg( 0, s );
  }

  /** Handles Acks based on the state we are currently in. */
  protected void handleAck( KermitPkt kPkt ) {
    // If we got this ack last time.  Toss it and move on
    if( lastAckedSeq == kPkt.getSeq() ) {
      dbg( 0, "Got duplicate Ack:"+lastAckedSeq );
      return;
    }

    // Ok - set our most recent seq ack to this ack.
    lastAckedSeq = kPkt.getSeq();
    curBadPacketCount = 0; // restart the retry count

    // Now - what do we do next
    if( kState == FH_PENDING ) { // should be the response for send_init
      // handle the send init response
      handleSendInit( kPkt );
      // start sending files
      sendNextFile();
    }
    else if( kState == ATTR_PENDING ) { // can we send the attribute packet
      byte[] packet = createAttributePacket();
      kState = DATA_PENDING;
      kStatus = "Sending attributes...";
      sendPacket( (byte)'A', packet );
    }
    else if( kState == OFF_PENDING ) { // can we send an EOT
      kStatus = "Complete";
      shutdown();
    }
    else if( kState == EOF_PENDING ) { // can we send an EOT
      sendEOF();
    }
    else if( kState == EOT_PENDING ) { // can we send an EOT
      // done with the last file.  Send the next file or send EOT
      sendNextFile();
    }
    else if( kState == DATA_PENDING ) { // can we send the next data packet?
      kStatus = "Sending data...";
      sendEncodedPacket( (byte)'D', kFileInputStream );
    }
  }

  protected void handleSendInit( KermitPkt kPkt ) {
    // process the information in the packet and send a new packet back.
    byte[] packet = kPkt.getBytes();
    //    dumpSendInit( packet );

    // process the information in the packet and send a new packet back.
    int i = kPkt.getDataIndex();

    if( i < packet.length ) sendMaxl = unByte( packet[i++] );
    sendMaxl = (sendMaxl > MAX_PACKET_LENGTH) ? MAX_PACKET_LENGTH : sendMaxl;

    if( i < packet.length ) sendTime = unByte( packet[i++] );
    i+= 2;	// No padding
    if( i < packet.length ) sendEol = (char)packet[i++];
    if( i < packet.length ) sendQctl = (char)packet[i++];

    // handle 8 bit byte settings
    if( i < packet.length ) {
      byte b = packet[i++];
      sendQbinChar = (char)b;
      if( ( b > 32 && b < 63 ) || (b > 95 && b < 127 ) ) {
	// they want parity, so pretend that we have it and handle 8bit shifts
	if( parity == 0 ) setParity( 1 ); //
	sendQbinFlag = true;
      } else if( b == 'Y' && parity != 0 ) {
	sendQbinFlag = true;
	sendQbinChar = '&';
      }
    }

    if( i < packet.length ) sendChkt = packet[i++] - '0';
    if( i < packet.length ) { // handle repeat character
      byte b = packet[i++];
      if( ( b > 32 && b < 63 ) || (b > 95 && b < 127 ) ) {
	sendRptChar = (char)b;
	sendRptFlag = true;
      }
    }

    // handle the capabilities
    if( i < packet.length ) {
      sendCapas = (byte)( unByte( packet[i++] ) & sendCapas );
    }

    /* This is a copy of the embedded kermit code.  As near as I can tell, the
     * the first of the four capability parameter bytes is going to be the 
     * window size.  The next two bytes will be the long packet length.
     *
     * Since the window size is even, we look for the first even packet.  This
     * allows other implementations to send extra CAPAS bytes, but only if
     * they are odd.  I don't really get it so I'm just trying to emulate it
     * and hope it works.
     */
    while( i < packet.length && ((unByte( packet[i] ) & 1) != 0 )) i++;

    // ok, first byte is the window byte.
    if( i < packet.length ) i++; // not yet supporting windows.

    if( ( sendCapas & CAP_LP ) > 0 && ((i+1) < packet.length) ) {
      // If this is an 'S' packet, we want to send back our largest packet
      // length no matter what they send us.  That way we let them know
      // what we are capable of.  Since we are only receiving, we will
      // never send back a big packet anyway.
      if( kPkt.getType() == 'Y' ) {
	sendLongMaxl = unByte(packet[i++]) * 95;
	sendLongMaxl += unByte(packet[i++]);
	sendLongMaxl = (sendLongMaxl >MAX_LONG_PACKET_LENGTH )
	  ? MAX_LONG_PACKET_LENGTH : sendLongMaxl;
      }
    }
  }

  protected void handleEOF( KermitPkt pkt ) {
    kStatus = "File completed";
    try {
      kFileOutputStream.close();
    } catch( Exception e ) {
      String s = "Could not close file transferred via kermit:"+e.getMessage();
      sendError( s );
      return;
    }
    kFileOutputStream = null;
    kState = RECV_FH_PENDING; // get ready for the next file header
    sendAck();
  }

  protected void handleEndOfTransmission( KermitPkt pkt ) {
    kFileOutputStream = null;
    sendAck();
    kStatus = "Completed";
    shutdown();
  }

  protected void handleFileHeader( KermitPkt pkt ) {
    byte[] packet = pkt.getBytes();

    // process the information in the packet and send a new packet back.
    kRemoteName = new String(packet, pkt.getDataIndex(), pkt.getDataLength());

    // let the controller check out this file name
    kFile = kController.kcGetFileToReceive( this );
    if( kFile == null ) {
      sendError( "Could not open " + kRemoteName+" locally" );
      return;
    }

    try {
      kFileOutputStream = new BufferedOutputStream
	( new FileOutputStream( kFile ) );
    } catch( Exception e ) {
      sendError( "Could not open "+kFile.getPath()+": "+e.getMessage() );
      return;
    }
    kStatus = "Waiting for attributes...";
    kState = RECV_DATA_PENDING;
    kByteProgress = 0;
    sendAck();
  }

  /** This is called with an input stream that can be appended to.  The
   * type is needed to keep track of the progress of sending the file.
   */
  protected void sendEncodedPacket( byte type, InputStream inputStream ) {
    // Until the packet is encoded, we don't know how big it is.
    // Until we know the size of the encoded packet we can't write
    // the header.
    //
    // therefore, let's build the encoded packet from as many inputBytes
    // as we can manage into a byteStream.  Then we'll create the byteStream
    // that will be used from the real packet.  Write the header bytes, add
    // the encoded bytestream and then figure the checksum.
    ByteArrayOutputStream dataStream = new ByteArrayOutputStream();

    int rptCount = 1;
    int lastByte = 0;
    int byteFromFile;
    int bytesRead = 0;

    // first, figure out the largest number of bytes we can send.

    // gotta have room for 5 bytes - the largest set of encoded characters
    // plus the checksum.  Since the sent packet will be the declared length
    // plus 2 (SOH,LEN), then we need to stop encoding when the packet size
    // is (sendMaxl - 3) or larger.
    //
    // we could optimize performance by checking to see if we have repeating
    // and 8 bit encoding turned on, but it's probably not worth it.

    // Figure out the largest packet size we can send.
    int largestPacketSize = sendMaxl - 2; // subtract 2; TYPE, SEQ
    if( (sendCapas & CAP_LP) > 0 && largestPacketSize > sendMaxl - 2 ) {
      largestPacketSize = sendLongMaxl; // with long pkt we only count chksum
    }
    largestPacketSize -= sendChkt;

    // make room for one last repeated, binary, prefixed char
    // this means that we will typically pack the packet with room
    // for 3 bytes to spare.

    while( dataStream.size() < largestPacketSize - 4 ) { 
      // read byte from file (buffered stream - not expensive)
      try {
	byteFromFile = inputStream.read();
      } catch( Exception e ) {
	// calling sendError could make this recursive except that when
	// sendEncodedPacket is called from sendError, there will be no io
	// error since the inputstream is from a byte array.
	sendError( "Problem reading " + kFile.getName() );
	return;
      }

      if( byteFromFile < 0 ) { // eof: prepare for eof, break
	kState = EOF_PENDING;
	kStatus = "Sending last packet...";
	break;
      }

      // We have a byte to write.

      // if we have repeat processing.  Process the repeat characters
      if( sendRptFlag ) {
	// do not do repeat processing if this is the first byte
	if( bytesRead > 0 ) {
	  if( lastByte != byteFromFile || rptCount >= 94 ) {
	    kermitEncodeByte( (byte)lastByte, dataStream, rptCount );
	    rptCount = 1;
	  } else {
	    rptCount++;
	  }
	}
	lastByte = byteFromFile;
      } // done with rpt processing
      else {
	kermitEncodeByte( (byte)byteFromFile, dataStream, 1 );
      } // done with non rpt processing

      bytesRead++;
    } // end while

    // if we hit EOF without reading any bytes, then send an EOF packet
    if( dataStream.size() <= 0 && type != 'E' ) {
      sendEOF();
      return;
    }

    // if we had repeat turned on, we still have to finish writing the chars
    // this is true even if we hit the EOF after reading some bytes.
    if( sendRptFlag ) {
      kermitEncodeByte( (byte)lastByte, dataStream, rptCount );
    }

    // Encode the data packet and send it.
    sendPacket( type, dataStream.toByteArray() );
    if( type != 'E' ) kByteProgress += bytesRead;
  }

  /** Encode one data byte to be included in a kermit packet */
  protected void kermitEncodeByte( byte b, ByteArrayOutputStream byteStream,
				   int cnt ) {
    // if we have a count of 3 or more, turn on repeating.
    if( cnt > 3 ) {
      byteStream.write( sendRptChar );
      byteStream.write( toByte( cnt ) );
      cnt = 1;
    }

    // check prefixed characters
    byte b7 = (byte)(b & 0177);
    byte b8 = (byte)(b & 0200);

    byte originalB = b;
    for( int i = 0; i < cnt; i++ ) {
      b = originalB;
      // send a byte with bit 8 turned on
      if( (b8 != 0 ) && sendQbinFlag ) {
	byteStream.write( sendQbinChar );
	b = (byte)(b - 0200);
      }

      // send a control character
      if( (b7&0177) < ' ' || (b7&0177) == 0177 ) { // control escape
	byteStream.write( sendQctl );      
	b = (byte)( b ^ 0100 );
      }
      else if( b7 == sendQctl ) { // control prefix
	byteStream.write( sendQctl );      
      }
      else if( ( b7 == sendQbinChar ) && sendQbinFlag ) { // binary prefix
	byteStream.write( sendQctl );
      }
      else if( ( b7 == sendRptChar ) && sendRptFlag ) { // repeat prefix
	byteStream.write( sendQctl );
      }

      byteStream.write( b );
    }
  }

  /** Decode the data portion of a kermit packet into 'data'. */
  ByteArrayOutputStream kermitDecodeBytes( byte[] bytes ) throws Exception {
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
    byte addBits = 0;
    int rptCount = 1;
    for( int i = 0; i < bytes.length; ) {
      byte b = bytes[i++];

      // get any repeat counter
      if( sendRptFlag && (sendRptChar == b) ) {
	rptCount = unByte( bytes[i++] );	// get the repeat count
	b = bytes[i++];				// set the repeat char
      }

      // handle binary prefix
      if( sendQbinFlag && b == sendQbinChar ) {
	addBits = (byte)0200;
	b = bytes[i++];
      }

      if( b == sendQctl ) {
	b = bytes[i++]; // grab next byte
	byte b7 = (byte)(b & 0177); // b7 should never be a prefix character
	if( ( b7 >= 0100 && b7 <= 0137 ) || b7 == '?' )
	  b = (byte)(b^0100); // process control char
      }
      appendToByteStream( byteStream, b, rptCount, addBits );
      rptCount = 1;
      addBits = 0;
    }
    return byteStream;
  }

  void appendToByteStream( ByteArrayOutputStream byteStream, byte b,
			   int rptCount, byte addBits ) {
    for( int i = 0; i < rptCount; i++ ) {
      byteStream.write( b+addBits );
    }
  }

  /** Can be used internally and externally.  Sends the packet, shuts the
   * kermit object down and interrupts the peer.  I don't know if this
   * is proper protocol - to send these extra characters.  Easily removed
   * if it doesnt.
   */
  public void sendError( String error ) {
    if( kState == OFF ) return; // only one error will be sent.
    sendEncodedPacket( (byte)'E', new ByteArrayInputStream(error.getBytes()) );
    kStatus = error;
    shutdown();
    interruptPeer();
  }

  protected void interruptPeer() {
    // send some control-C's to interrupt kermit if it hasn't gotten the msg
    kController.kcWrite( this, new byte[] {3,3} );
  }

  /** Close all file streams and put kermit in an off state. */
  public void shutdown() {
    kState = OFF;
    kLastPkt = null;

    if( kFileInputStream != null ) {
      try {
	kFileInputStream.close();
      } catch( Exception e ) {}
    }

    if( kFileOutputStream != null ) {
      try {
	kFileOutputStream.close();
      } catch( Exception e ) {}
    }

    // This should be superfluous - we should be going away.
    kFileInputStream = null;
    kFileOutputStream = null;
  }

  protected void handleData( KermitPkt pkt ) {
    // process the information in the packet and send a new packet back.
    kStatus = "Receiving Data...";
    kState = RECEIVING_DATA;
    int dataIndex = pkt.getDataIndex();
    int dataLength = pkt.getDataLength();
    byte[] packet = pkt.getBytes();
    byte[] data = new byte[dataLength];
    for( int i = dataIndex; (i - dataIndex) < dataLength; i++ ) {
      data[i-dataIndex] = packet[i];
    }
    try {
      byte[] byteArray = kermitDecodeBytes( data ).toByteArray();
      kByteProgress += byteArray.length;
      kFileOutputStream.write( byteArray );
    } catch( Exception e ) {
      sendError( e.getMessage() );
      return;
    }
    sendAck();
  }

  protected byte[] createAttributePacket() {
    /* We will send
     *	type (binary unless otherwise spec'd)
     *  file date last modfied
     *	size in KBytes
     *	size in bytes
     */
    // Make it simpler by using bytearray
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
    byte[] bString;
    // type
    byteStream.write( '"' );
    byteStream.write( toByte( 2 ) );
    bString = (new String( "B8" )).getBytes();
    byteStream.write( bString, 0, bString.length );

    // size in KB
    byteStream.write( '!' );
    long fileLength = kFile.length();
    String s = Long.toString( fileLength/1024 );
    byteStream.write( toByte( s.length() ) );
    bString = s.getBytes();
    byteStream.write( bString, 0, bString.length );

    // file creation date
    byteStream.write( '#' );
    s = dateFormatter.format( new Date( kFile.lastModified() ) );
    byteStream.write( toByte( s.length() ) );
    bString = s.getBytes();
    byteStream.write( bString, 0, bString.length );

    // size in Bytes
    byteStream.write( '1' );
    s = Long.toString( fileLength );
    byteStream.write( toByte( s.length() ) );
    bString = s.getBytes();
    byteStream.write( bString, 0, bString.length );

    byteStream.write( ' ' ); // placeholder
    return( byteStream.toByteArray() );
  }

  protected byte[] createSendInitPacket() {
    byte[] packet = new byte[13];
    int i = 0;
    packet[ i++ ] = toByte( sendMaxl );
    packet[ i++ ] = toByte( sendTime );
    packet[ i++ ] = toByte( sendNpad );
    packet[ i++ ] = (byte)sendPadc;
    packet[ i++ ] = (byte)sendEol;
    packet[ i++ ] = (byte)sendQctl;
    if( parity == 0 ) {
      packet[ i++ ] = (byte)'Y'; // turn it off, but allow it
    } else {
      packet[ i++ ] = (byte)sendQbinChar;
    }
    packet[ i++ ] = (byte)(sendChkt + '0');
    packet[ i++ ] = (byte)sendRptChar;
    packet[ i++ ] = toByte( (int)sendCapas );
    packet[ i++ ] = toByte( 0 ); // not supporting windows
    packet[ i++ ] = toByte( sendLongMaxl / 95 ); // big length byte 
    packet[ i++ ] = toByte( sendLongMaxl % 95 ); // small length byte
    return packet;
  }

  protected void handleAttributes( KermitPkt pkt ) {
    byte[] packet = pkt.getBytes();
    int endIndex = pkt.getDataIndex() + pkt.getDataLength();
    int fileSizeInBytes = 0;
    int i = 4;
    while( i < endIndex ) {
      StringBuffer buf = new StringBuffer();
      byte attrType = packet[i++];
      int attrLen = unByte( packet[i++] );
      for( int j = 0; j < attrLen; j++ ) buf.append( (char)packet[i++] );
      String bufStr = buf.toString();
      switch( attrType ) {
      case '!':		/* File length in K */
      case '"':		/* File type */
	if( attrType == '!' ) {		/* Length in KBytes */
	  try {
	    if( fileSizeInBytes == 0 )
	      fileSizeInBytes = Integer.parseInt( bufStr ) * 1024;
	  } catch( Exception e ) {
	    dbg( 0, "parse of int failed:" + bufStr );
	  }
	} else {				/* Type */
	  if( bufStr.charAt( 0 ) == 'A' )	/* Text */
	    {}
	  else if( bufStr.charAt( 0 ) == 'B')	/* Binary */
	    {}
	}
	break;
      case '#':		/* File creation date */
	break;

      case '1':         /* File length in bytes */
	try {
	  fileSizeInBytes = Integer.parseInt( bufStr );
	} catch( Exception e ) {
	  dbg( 0, "sizeInBytes failed:" + e );
	}
	break;

      default:
	dbg( 0, "Unknown flag:" + (char)attrType + ":" + bufStr );
      }
    }
    sendAck();
    kFileSize = fileSizeInBytes;
  }

  protected void handleError( KermitPkt pkt ) {
    String s = new String( pkt.getBytes(), pkt.getDataIndex(),
			   pkt.getDataLength() );
    kStatus = "Kermit Error:" + s;
    shutdown();
  }

  /** These are synchronized primarily to keep Error packets, which may
   * be sent asynchronously, from interfering with the packet currently
   * being sent.
   */
  protected synchronized void writePacket( byte[] packet ) {
    kController.kcWrite( this, packet );
    kController.kcWrite( this, new byte[] { (byte)unByte( (byte)sendEol ) } );
  }

  protected synchronized void sendPacket( byte type, byte[] data ) {
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
    byte[] packet;
    byteStream.write( sendSOH );
    int declaredLength;
    int length;

    // first figure out the length
    if( (data.length + sendChkt + 2) > sendMaxl ) { // long packet
      declaredLength = 0;
      length = data.length + sendChkt;
    } else {
      length = declaredLength = data.length + sendChkt + 2;
    }

    byteStream.write( toByte( declaredLength ) );
    byteStream.write( toByte( kSeq ) );
    byteStream.write( type );

    // Tracks seqNo for sending packets
    if( kState >= SEND_PENDING ) kSeq = getNextSeq( kSeq );

    if( declaredLength == 0 ) { // long packet
      byteStream.write( toByte( length / 95 ) );
      byteStream.write( toByte( length % 95 ) );
      packet = byteStream.toByteArray();
      byteStream.write( toByte( checksum1( packet, 1, 6 ) ) );
    } else {
    }

    // add dataStream
    byteStream.write( data, 0, data.length );
    for( int i = 0; i < sendChkt; i++ ) byteStream.write( 0 );

    packet = byteStream.toByteArray();

    //packet[packet.length-1] = toByte( checksum1(packet, 1, packet.length-1));

    // now add the checksum into the last three bytes
    int chk = 0;
    int i = packet.length - sendChkt;
    switch( sendChkt ) {                   /* Add block check */
    case 1:                           /* 1 = 6-bit chksum */
      packet[i++] = toByte( checksum1( packet, 1, i ) );
      break;
    case 2:                           /* 2 = 12-bit chksum */
      chk = checksum2( packet, 1, i );
      packet[i++] = toByte( (chk >> 6) & 077);
      packet[i++] = toByte( chk & 077);
      break;
    case 3:                           /* 3 = 16-bit CRC */
      chk = checksum3( packet, i, 1 );
      packet[i++] = toByte( (chk & 0170000) >> 12);
      packet[i++] = toByte( (chk >> 6) & 077);
      packet[i++] = toByte( chk & 077 );
      break;
    }

    writePacket( packet );
    kLastPkt = packet; // save this packet in case we have to retransmit
  }

  void sendEOF() {
    kStatus = "Completed sending file " + kFile.getName() + "...";
    sendPacket( (byte)'Z', new byte[0] );
    try {
      kFileInputStream.close();
    } catch( Exception e ) {
      dbg( 0, "Could not close file "+kFile.getPath()+": "+e.getMessage() );
    }
    kFileInputStream = null;
    kFile = null;
    kState = EOT_PENDING;
  }

  protected void sendEOT() {
    kStatus = "Finishing session...";
    sendPacket( (byte)'B', new byte[0] );
    kState = OFF_PENDING;
  }

  protected void sendAck() {
    curBadPacketCount = 0; // restart the retry count
    sendPacket( (byte)'Y', new byte[0] );
  }

  protected void sendNak( byte seq ) {
    if( isBadPacketCountTooHigh( "Unexpected response" )) return;
    sendPacket( (byte)'N', new byte[0] );
  }

  protected void handleNak( KermitPkt pkt ) {
    if( kState == OFF || kState == RECV_PENDING ) return; // not ready yet

    if( isBadPacketCountTooHigh( "got NAK" )) return;
    int seqNo = pkt.getSeq();
    /* This NAK indicates that the receiver needs the packet specified
     * by the seq in this packet.  This could be because they got a packet
     * where the cksum didn't match, waited for a packet and got none or
     * got a packet with an unexpected seq no
     */
    if( kLastPkt == null ) {
      // we happened to just turn kermit on and we got this before we got
      // a chance to send out our first packet OR, we are inadvertently
      // in receive state and our peer is also in receive state sending
      // NAKS.  Easiest thing is to just ignore this NAK.
      return;
    }

    int lastSentSeq = unByte( kLastPkt[2] );

    // If the seq they ask for is the last one we sent, just send it.
    if( lastSentSeq == pkt.getSeq() ) {
      writePacket( kLastPkt );
      return;
    }

    /* If the seq they ask for is one more than the last one we sent, that
     * means they got the last packet and we didn't get their corresponding
     * ACK.  We need to treat this as if it were the ACK to the last packet
     * that we never got.
     */
    if( getNextSeq( lastSentSeq ) == pkt.getSeq() ) {
      if( kState == FH_PENDING ) { // we never got the send_init reply
	sendError( "Sender never got the send_init reply" );
	return;
      }
      // turn this packet into an ACK and handle it.
      handleAck( new KermitPkt( lastSentSeq, (byte)'Y', pkt.getLength(),
				pkt.getBytes() ) );
      return;
    }
    
    // We cannot handle this request
    sendError( "Receiver is asking for packet " + pkt.getSeq()
	       + " but we only have packet " + lastSentSeq + " to resend" );
  }

  /* These get the current state of kermit */
  /** Returns the name of the file we are sending or receiving. */
  public String getRemoteFileName() { return kRemoteName; }
  /** Returns the local File object of the file we are sending or receiving. */
  public File getLocalFile() { return kFile; }
  /** Returns the latest news on kermit - for GUI display */
  public String getStatus() { return kStatus; }
  /** Returns the latest state of kermit */
  public int getState() { return kState; }
  /** Returns the size of the file we are getting or sending if available */
  public int getFileSize() { return kFileSize; }
  /** Returns the number of bytes we have sent or received */
  public int getByteProgress() { return kByteProgress; }
  /** Returns the time we should wait for a timeout a*/
  public int getTimeout() { return sendTime; }

  protected int getNextSeq( int num ) { return (num >=64)?0:num+1; }

  public String toString() {
    return "kermit:"
      +"From:"+kRemoteName+",status("+kStatus+"),"
      +"fileSize("+kFileSize+"),state("+kState+"),"
      +"byteProgress("+kByteProgress+"),timeout("+sendTime+"),"
      +"seq("+kSeq+"))";
  }

  /* This embedded class is used to collect bytes and form a kermit packet on
   * the receiving side.  It is not used on the send side at all.
   */
  protected class KermitPkt {
    final int EXPECTING_LENGTH = 1;
    final int EXPECTING_SEQ = 2;
    final int EXPECTING_TYPE = 3;
    final int EXPECTING_DATA = 4;
    final int EXPECTING_CHECKSUM = 5;
    final int PACKET_COMPLETE = 6;

    int		length = -1;	// also indicates long packet with a 0
    int		seq = -1;
    byte	type = -1;
    int		currentIndex = 0;
    int		dataLength = -1;
    int		dataIndex = -1;
    byte	chkSumLength = 1;
    boolean	packetComplete = false;
    byte[]	bytes;

    boolean longPacket = false;
    int state = EXPECTING_LENGTH;

    KermitPkt( int seq, byte type, int length, byte[] packet ) {
      this.seq = seq;
      this.type = type;
      this.length = length;
      if( length <= 94 ) { // 1 byte len
	dataLength = length - (chkSumLength + 2);
	dataIndex = 4;
	bytes = new byte[length+2]; // length = length + byte0 and byte1
      } else { // long packet
	dataLength = length - chkSumLength;
	dataIndex = 7;
	bytes = new byte[length+7]; // length = length + all ext header bytes
      }
      // this copies the header bytes of the packet.
      for( int i = 0; i < dataIndex; i++ ) bytes[i] = packet[i];
      currentIndex = dataIndex;
    }

    void addByte( byte b ) {
      if( currentIndex < bytes.length ) // data
	bytes[currentIndex++] = b;
      else {
	dbg( 0, "Extraneous byte in packet - ignoring " + b );
      }
      if( currentIndex >= bytes.length ) packetComplete = true;
    }

    boolean isPacketOk() {
      if( type == 'S' || kState == FH_PENDING ) return isPacketOk( 1 );
      return isPacketOk( sendChkt );
    }
    
    boolean isPacketOk( int chkType ) {
      int thisSum;
      byte checkSum;
      switch( chkType ) {
      case 1:
	checkSum = toByte( checksum1( bytes, 1, bytes.length-1 ) );
	if( checkSum == bytes[bytes.length - 1] ) return true;
	break;
      case 2:
	thisSum =
	  (unByte( bytes[bytes.length - 2] ) << 6)
	  | unByte( bytes[bytes.length - 1] );
	if( thisSum == checksum2( bytes, 1, bytes.length-2 ) ) return true;
	// let error packets have 1 byte checksums
	if( type == 'E' ) return isPacketOk( 1 );
	break;
      case 3:
	thisSum =
	  (unByte( bytes[bytes.length - 3] ) << 12)
	  | (unByte( bytes[bytes.length - 2] ) << 6)
	  | unByte( bytes[bytes.length - 1] );
	if( thisSum == checksum3( bytes, 1, bytes.length-3 ) ) return true;
	// let error packets have 1 byte checksums
	if( type == 'E' ) return isPacketOk( 1 );
	break;
      }
      // needs to be adjusted for other checksum types
      dbg( 0, "checksum mismatch" );
      return false;
    }

    boolean isComplete() { return packetComplete; }
    byte[] getBytes() { return bytes; }
    int getSize() { return bytes.length; } // size of real packet
    int getLength() { return length; } // length declared by peer
    int getSeq() { return seq; } // seq declared by peer
    byte getType() { return type; } // type declared by peer
    int getDataIndex() { return dataIndex; }
    int getDataLength() { return dataLength; }
  }

  protected void dbg( int _debug, Object msg ) {
    if( _debug > debug ) return;
    System.err.println( getClass().getName() + ":" + msg );
  }
}
